# test_integration_030_get.py
from http import server
import pytest
# requires pytest_dependency to be installed
import unittest
import responses
from fastapi.testclient import TestClient
import requests
import re
import logging
import sys
import os
import json
from pathlib import Path

# so the directly in which you run pytest or unittest doesn't matter as much
sys.path.append(os.path.abspath('..'))
sys.path.append('tests')

# local
from main import app
from auth import auth
import conf
import testlib

# override dependency injection for authentication to entitlement service
from override_depends import override_require_authorized_user, set_authorize_session, ADMIN_ONLY_SVC, USER_AND_ADMIN_SVC, USER_ONLY_SVC, OTHER_ONLY_SVC

TEST_DATA_DIR = Path(__file__).resolve().parent / 'templates'

client = TestClient(app)

@pytest.mark.dependency(name="require_token")
def test_require_token(token):
    assert token is not None, "No token provided on command line"

@pytest.mark.opa_access
@pytest.mark.dependency(depends=["require_token"])
def test_local_fetch_policies(token, data_partition):
    """
    Check /policies API endpoint is available
    """
        
    r = client.get(conf.SERVICE_BASE_PATH + "/policies",
        headers={'Authorization': 'Bearer ' + token,
        'data-partition-id': data_partition})
    print(r.text)
    assert r.status_code == 200, f"Expected to get 200 from /policies {r.text}"

@pytest.mark.dependency(depends=["require_token"])
def test_serviceurl_fetch_policies(token, data_partition, service_url):
    """
    Check /policies API endpoint is available
    """

    r = requests.get(service_url + conf.SERVICE_BASE_PATH + "/policies",
        headers={'Authorization': 'Bearer ' + token,
        'data-partition-id': data_partition})
    print(r.text)
    assert r.status_code == 200, f"Expected to get 200 from /policies {r.text}"

@pytest.mark.dependency(depends=["require_token"])
def test_fetch_policies_unknown_data_partition_id(token):
    """
    Check /policies API endpoint using a bad partition id
    Should always be unauthorized 401 even with a good token
    """

    if conf.MOCK_ENTITLEMENT:
        pytest.skip("MOCK ENTITLEMENT ENABLED")

    r = client.get(conf.SERVICE_BASE_PATH + "/policies",
        headers={'Authorization': 'Bearer ' + token,
        'data-partition-id': "unknownpartitionid"})
    print(token)
    assert r.status_code == 401, f"expect unauthenticated response {r.status_code} {r.text}"

@pytest.mark.opa_access
@pytest.mark.mock
@pytest.mark.dependency(depends=["require_token"])
def test_local_instance_policy_dataauthz_mock(token, data_partition):
    get_instance_policy(token=token, data_partition=data_partition, policy_id="dataauthz.rego")

@pytest.mark.dependency(depends=["require_token"])
def test_instance_policy_dataauthz_service_url(token, data_partition, service_url):
    get_instance_policy(token=token, data_partition=data_partition, policy_id="dataauthz.rego", service_url=service_url)

@pytest.mark.opa_access
@pytest.mark.mock
@pytest.mark.dependency(depends=["require_token"])
def test_local_instance_policy_entitlements_mock(token, data_partition):
    get_instance_policy(token=token, data_partition=data_partition, policy_id="entitlements.rego")

@pytest.mark.dependency(depends=["require_token"])
def test_instance_policy_entitlements_sevice_url(token, data_partition, service_url):
    get_instance_policy(token=token, data_partition=data_partition, policy_id="entitlements.rego", service_url=service_url)

@pytest.mark.opa_access
@pytest.mark.mock
@pytest.mark.dependency(depends=["require_token"])
def test_local_instance_policy_legal_mock(token, data_partition, service_url):
    get_instance_policy(token=token, data_partition=data_partition, policy_id="legal.rego")

@pytest.mark.dependency(depends=["require_token"])
def test_instance_policy_legal_service_url(token, data_partition, service_url):
    get_instance_policy(token=token, data_partition=data_partition, policy_id="legal.rego", service_url=service_url)

def get_instance_policy(token, data_partition, policy_id, service_url=None):
    """
    Check /policies/osdu/instance/ endpoint
    requires dataauthz.rego to be in instance
    """
    url = conf.SERVICE_BASE_PATH + "/policies/osdu/instance/" + policy_id
    policy_id_short = policy_id.removesuffix('.rego')
    if service_url:
        url = service_url + url 
        print(url)
        r = requests.get(url,
            headers={'Authorization': 'Bearer ' + token,
            'data-partition-id': data_partition})
    else: # mock aka no CLOUD_PROVIDER
        print(url)
        r = client.get(url,
            headers={'Authorization': 'Bearer ' + token,
            'data-partition-id': data_partition})
    assert r.status_code == 200, f"Get instance policy {policy_id} failed. Load policies before running test. AdminCLI 'make load_data'"
    assert "result" in r.text, f"expect result for {policy_id}"
    if service_url:
        assert f"osdu/instance/{policy_id}" in r.text, f"expect result for {policy_id}"
        assert f"package osdu.instance.{policy_id_short}" in r.text, f"expect result for {policy_id}"
    
@pytest.mark.opa_access
@pytest.mark.mock
@pytest.mark.dependency(depends=["require_token"])
def test_get_diag_policies_mock(token, data_partition):
    """
    get diagnostic policies mock
    """
    if conf.ENABLE_DEV_DIAGNOSTICS == False:
        pytest.skip("Skipping diagnostic test")
        
    num_tests = 0
    r = client.get("/diag/policies",
        headers={'Authorization': 'Bearer ' + token,
        'data-partition-id': data_partition})
    assert "result" in r.text, f"result not expected {r.text}"
    #print(f"text: {r.text}",file=sys.stderr)
    assert r.status_code == 200, "policies Page is available"
    p = os.path.dirname(os.path.abspath(__file__))
    template_datadir = os.path.join(p, '..', 'templates')
    for filename in sorted(os.listdir(template_datadir)):
        filetype='.rego'
        if filename.endswith(filetype):
            if filename not in r.text:
                json_str = json.dumps(r.json(), indent=4, sort_keys=True)
                with open("policy_dump_debug.json", "w") as outfile:
                    outfile.write(json_str)
            assert filename in r.text, f"result not expected data from filename {filename} not in OPA. See policy_dump_debug.json" 
            num_tests = num_tests + 1
    assert num_tests >= 7, "At least 7 polices were tested"

@pytest.mark.dependency(depends=["require_token"])
def test_get_diag_policies_service_url(token, data_partition, service_url):
    """
    get diagnostic policies
    And make sure at least 7 policies are retrieved and all the policies in template directory were loaded
    """
    if conf.ENABLE_DEV_DIAGNOSTICS == False:
        pytest.skip("Skipping diagnostic test")
        
    num_tests = 0
    r = requests.get(service_url + "/diag/policies",
        headers={'Authorization': 'Bearer ' + token,
        'data-partition-id': data_partition})

    assert "result" in r.text, f"result not expected {r.text}"
    #print(f"text: {r.text}",file=sys.stderr)
    assert r.status_code == 200, "diag policies API is not available"
    p = os.path.dirname(os.path.abspath(__file__))
    template_datadir = os.path.join(p, '..', 'templates')
    for filename in sorted(os.listdir(template_datadir)):
        filetype='.rego'
        if filename.endswith(filetype):
            assert filename in r.text, f"result not expected data from filename {filename} not in OPA: {r.text}" 
            num_tests = num_tests + 1
    assert num_tests >= 7, "At least 7 polices were tested"

# @pytest.mark.dependency(depends=["require_token"])
# def test_fetch_policies_advanced(token, data_partition, service_url):
#     """
#     get partition policies
#     Test get /policies/osdu/partition/<data_partition>/<policy> using built-in
#     And make sure at least 7 policies are retrieved and all the policies in template directory were loaded
#     """
#     num_tests = 0
#     p = os.path.dirname(os.path.abspath(__file__))
#     template_datadir = os.path.join(p, '..', 'templates')
#     for filename in sorted(os.listdir(template_datadir)):
#         filetype='.rego'
#         if filename.endswith(filetype):
#             id = f"osdu/partition/{data_partition}/{filename}"
#             r = requests.get(service_url + conf.SERVICE_BASE_PATH + "/policies/" + id,
#                 headers={'Authorization': 'Bearer ' + token,
#                 'data-partition-id': data_partition})
#             assert r.status_code == 200, f"Fetch of {id}: {r.text}"
#             assert "result" in r.text, f"result not expected {r.text}"
#             num_tests = num_tests + 1
#     assert num_tests >= 7, "At least 7 polices were tested"

@pytest.mark.dependency(depends=["require_token"])
@pytest.mark.parametrize("regofile", testlib.get_rego_templates())
def test_get_policy(token, data_partition, service_url, domain, regofile):
    assert token is not None, "No token provided on command line"
    filetype='.rego'
    if regofile.endswith(filetype):
        id = f"osdu/partition/{data_partition}/{regofile}"
        r = requests.get(service_url + conf.SERVICE_BASE_PATH + "/policies/" + id,
            headers={'Authorization': 'Bearer ' + token,
            'data-partition-id': data_partition})
        assert r.status_code == 200, f"Fetch of {id}: {r.text}"
        assert "result" in r.text, f"result not expected {r.text}"

@pytest.mark.dependency(depends=["require_token"])
def test_get_osdu_partition_policies_serviceurl(token, data_partition, service_url):
    """
    Test get /policies/osdu/partition/<data_partition>/<policy>
    """
    get_osdu_partition_policies(token=token,
        data_partition=data_partition,
        service_url=service_url)

@pytest.mark.opa_access
@pytest.mark.dependency(depends=["require_token"])
def test_get_osdu_partition_policies_local(token, data_partition):
    get_osdu_partition_policies(token=token,
        data_partition=data_partition)

def get_osdu_partition_policies(token, data_partition, service_url=False):
    num_tests = 0
    p = os.path.dirname(os.path.abspath(__file__))
    template_datadir = os.path.join(p, '..', 'templates')
    for filename in sorted(os.listdir(template_datadir)):
        filetype='.rego'
        if filename.endswith(filetype):
            id = f"osdu/partition/{data_partition}/{filename}"
            policy_id_short = filename.removesuffix('.rego')
            if service_url:
                url = service_url + conf.SERVICE_BASE_PATH + "/policies/" + id

                r = requests.get(url,
                    headers={'Authorization': 'Bearer ' + token,
                    'data-partition-id': data_partition})
            else: # mock
                url = conf.SERVICE_BASE_PATH + "/policies/" + id
                r = client.get(url,
                    headers={'Authorization': 'Bearer ' + token,
                    'data-partition-id': data_partition})
            assert r.status_code == 200, f"Fetch of {filename} {url}: {r.text}"
            assert "result" in r.text, f"result not expected {r.text} {url}"
            
            if service_url:
                assert f"osdu/partition/{data_partition}/{filename}" in r.text, f"expect result for {filename} osdu/partition/<dp>/<policy.rego>"
                assert f"package osdu.partition[\\\"{data_partition}\\\"].{policy_id_short}" in r.text, f"expect result for {filename}"
            print(r.text)
            num_tests = num_tests + 1
    assert num_tests >= 7, "At least 7 polices were tested"
